/**
 * perf.js
 * Some performance tests.
 */

var Vec2D = require('./vec2d.js'),
  program = require('commander');

program.version('0.0.1').option('-c, --count <count>', 'Number of times to perform each operation', Number, 100000).option('-rmin, --rmin <rmin>', 'Smallest random number to use.', Number, -10000).option('-rmax, --rmax <rmax>', 'Largest random number to use.', Number, 10000).parse(process.argv);

var RAND_MIN = program.rmin,
  RAND_MAX = program.rmax,
  OPERATION_COUNT = program.count;


var outputData = {};

function randomFloat(min, max) {
  if (!min) {
    min = RAND_MIN;
  }

  if (!max) {
    max = RAND_MAX;
  }

  return min + (max - min) * Math.random();
}

function randomInt(min, max) {
  if (!min) {
    min = RAND_MIN;
  }

  if (!max) {
    max = RAND_MAX;
  }

  return Math.floor(min + (1 + max - min) * Math.random());
}

var util = {
  start: null,
  maxTime: 0,

  startTimer: function() {
    this.start = new Date().getTime();
  },

  getTimePassed: function() {
    var time = new Date().getTime() - this.start;

    if (time > this.maxTime) {
      this.maxTime = time;
    }
  },

  runTask: function(name, callback) {
    var self = this;

    suite[name](function() {
      self.startTimer();
      console.log('\nStart task: ' + name);
    }, function() {
      var time = self.getTimePassed().toString();
      console.log('Finished task: ' + name + ' in ' + time + 'ms');
      if (callback) {
        callback();
      }
      return time;
    });
  }
};

var suite = {
  generate: function(start, end) {
    function generateFloat32(start, end) {
      start();
      var i = OPERATION_COUNT;

      Vec2D.useFloat32Arrays();

      var float32Vectors = [];
      while (i) {
        float32Vectors.push(Vec2D.create(randomFloat(), randomFloat()));
        i--;
      }

      return end();
    }

    function generateStandard(start, end) {
      start();
      var i = OPERATION_COUNT;

      Vec2D.useStandardArrays();

      var vectors = [];
      while (i) {
        vectors.push(Vec2D.create(randomFloat(), randomFloat()));
        i--;
      }

      return end();
    }


    function generateObjects(start, end) {
      start();
      var i = OPERATION_COUNT;

      Vec2D.useObjects();

      var vectors = [];
      while (i) {
        vectors.push(Vec2D.create(randomFloat(), randomFloat()));
        i--;
      }

      return end();
    }

    return {
      'Float32': generateFloat32(),
      'Array': generateStandard(),
      'Object': generateObjects()
    };
  },


  add: function() {
    function addFloat32(start, end) {
    var vectors = this.generateFloat32(function() {}, function() {});
    start();

    var i = vectors.length - 1;
    while (i) {
      vectors[i].add(vectors[i]);
      i--;
    }

    end();
  },

  function addArray(start, end) {
    var vectors = this.generateStandard(function() {}, function() {});
    start();

    var i = vectors.length - 1;
    while (i) {
      vectors[i].add(vectors[i]);
      i--;
    }

    end();
  },

  function addObjects(start, end) {
    var vectors = this.generateObjects(function() {}, function() {});
    start();

    var i = vectors.length - 1;
    while (i) {
      vectors[i].add(vectors[i]);
      i--;
    }

    end();
  },
  }
  

  updateFloat32: function(start, end) {
    var vectors = this.generateFloat32(function() {}, function() {});
    start();

    var i = vectors.length - 1;
    while (i) {
      vectors[i].setAxes(randomFloat(), randomFloat());
      i--;
    }

    end();
  },

  updateArray: function(start, end) {
    var vectors = this.generateStandard(function() {}, function() {});
    start();

    var i = vectors.length - 1;
    while (i) {
      vectors[i].setAxes(randomFloat(), randomFloat());
      i--;
    }

    end();
  },

  updateObjects: function(start, end) {
    var vectors = this.generateObjects(function() {}, function() {});
    start();

    var i = vectors.length - 1;
    while (i) {
      vectors[i].setAxes(randomFloat(), randomFloat());
      i--;
    }

    end();
  }
}

console.log('Running performance tests with ' + OPERATION_COUNT + ' vectors per test item...');
outputData['create'] = util.runTask('generate');
util.runTask('addFloat32');
util.runTask('addArray');
util.runTask('addObjects');
util.runTask('updateFloat32');
util.runTask('updateArray');
util.runTask('updateObjects');